#!/usr/bin/env python3
# vi:noet:

import re
import copy
import os
import tempfile

import networkx as nx

import pypint

def get_edge_color(g,u,v):
	return g.edge[u][v]['color']

def write_muted_automata(nb_muts, permanent, inputfile, exclude, only, outputfile=None):

	an = pypint.load(inputfile)				#load the file

	if outputfile is None:
		fd, outputfile = tempfile.mkstemp(suffix="muted_automata_network.an")
		f = os.fdopen(fd, "w")
	else:
		f = open(outputfile, "w")

	initial_state = an.info["initial_state"]
	initial_state["x"]=0
	initial_state["z"]=0

	f.write("z [0,1]\n")				#z is 0 or 1
	f.write("x [%s] (* %s mutations allowed *)\n" % \
	    (",".join(map(str, range(0,nb_muts))), nb_muts))
	f.write("\n(* reset mutations *)\n")
	for i in range(0, nb_muts-1):
		f.write("{z 1 -> 0; x "+str(i)+" -> "+str(i+1)+"}\n")	#mutation "reset"
	f.write("\n")

	# copying the other states
	for a, ls in an.local_states.items():
		f.write("%s [%s]\n" % (a,",".join(map(str,ls))))
	f.write("\n")

	mut_automata = an.automata if not only else only
	if exclude:
	    mut_automata = [a for a in mut_automata if a not in exclude]

	# if permanent mutations, adding locks
	if permanent:
		for a in mut_automata:
			f.write("lock_%s [0,1]\n" % a)
			initial_state["lock_%s" % a]=0

	# adding transitions, slightly different if permanent or not:
	for tr in an.local_transitions:
	    if permanent:
		    for a in tr.modified_automata:
			    if a in mut_automata:
				    tr.conds["lock_%s" % a] = 0
	    f.write("%s\n" % tr)
	f.write("\n")

	# adding locking of the value transitions (mutations)
	# adding mutations that change the value. If permanent, the value is the locked
	for a in mut_automata:
		if permanent:
			f.write("{z 0 -> 1; lock_%s 0 -> 1}\n" % a)
		ls = an.local_states[a]
		for (i,j) in [(i,j) for i in ls for j in ls if i != j]:
			if permanent:
				f.write("{z 0 -> 1; %s %d -> %d; lock_%s 0 -> 1}\n" % (a,i,j,a))
			else:
				f.write("{z 0 -> 1; %s %d -> %d}\n" % (a,i,j))

	f.write("\ninitial_state %s\n" % an.initial_state.to_pint(full=True))
	f.close()

	# returns the name of the new an file and the modified initial state
	return (outputfile, initial_state)


def dict_matches(main, pattern):
	for k,v in pattern.items():
		if main[k] != v:
			return False
	return True

def parse_localstate(ls):
	a,i = ls.split("=")
	return (a.strip(), int(i))

def parse_label(label):
	return dict([parse_localstate(ls) for ls in label.split(",")])

def parse_perturbation_stg(an, g, initial_state, final_state):
	"""
	1. parse node state from label
	2. color red and label perturbation edges
	3. compress z 1 -> 0 transitions
	4. set x = z+x, del z
	5. find matching initial and final states
	"""

	# automata of interest (i.e., not x or z)
	A = [a for a in an.automata if a not in ["z", "x"]]

	def transform_to_red_arrow(u,v):
		g.edge[u][v]['color']="red"
		lu = g.node[u]["state"]
		lv = g.node[v]["state"]
		def to_automata(a):
			return a[5:] if a.startswith("lock_") else a
		diff = set([to_automata(a) for a in A if lu[a] != lv[a]])
		assert len(diff) == 1, diff
		label = ", ".join(["%s=%d" % (a,lv[a]) for a in diff])
		g.edge[u][v]["label"] = label

	for n in g.node:
		g.node[n]["state"] = parse_label(g.node[n]["label"])

	nodes_to_remove = set()
	forward_edges = []
	for (u,v) in g.out_edges():
		if g.node[u]["state"]["z"] < g.node[v]["state"]["z"]:
			transform_to_red_arrow(u,v)
		elif g.node[u]["state"]["z"] > g.node[v]["state"]["z"]:
			nodes_to_remove.add(v)
			forward_edges += [(u,v,w) for (_,w) in g.out_edges(nbunch=v)]
	for (u,v,w) in forward_edges:
		g.add_edge(u, w, g.edge[v][w])
	for n in nodes_to_remove:
		g.remove_node(n)

	#finding initial state and reducing the labels (from "x=a, z=b" to "x=a+b")
	i_states=[]
	ends = []
	for n, data in g.nodes(data=True):
		if dict_matches(data["state"], initial_state):
		    i_states.append(n)
		elif dict_matches(data["state"], final_state):
		    ends.append(n)
		data["state"]["x"] += data["state"]["z"]
		del data["state"]["z"]
		data["label"] = ",".join(["%s=%d" % kv for kv in data["state"].items()])
	assert len(i_states) == 1
	return i_states, ends


def dump_graph(g, outfile, begins, ends, show_label=True):
	with open(outfile, "w") as f:
		f.write("digraph {\n")
		for (u,v,data) in g.edges(data=True):
			f.write("%s -> %s" % (u,v))
			if g.edge[u][v].get("color") == "red":
				f.write("[color=red,label=\"%s\"]" % g.edge[u][v]["label"])
			f.write(";\n")
		if show_label:
			for n, data in g.nodes(data=True):
				if "label" in data:
					f.write(" %s[label=\"%s\"" % (n, data["label"]))
				if n in begins:
					f.write(",color=blue,style=filled")
				if n in ends:
					f.write(",color=green,style=filled")
				f.write("];\n")
		else:
			for n in begins:
				f.write("	%s[color=blue,style=filled];\n" % n)
			for n in ends:
				f.write("	%s[color=green,style=filled];\n" % n)
		f.write("}\n")


def graph_wrap(g):
	sccg = nx.condensation(g)
	print("%d nodes in condensed graph" % len(sccg))
	n = 0
	for (u,v,data) in g.edges(data=True):
		if "color" in data:
			n += 1
			mu = sccg.graph['mapping'][u]
			mv = sccg.graph['mapping'][v]
			e = sccg.edge[mu][mv]
			if "color" not in e:
				e["color"] = data["color"]
				e["label"] = set()
			e["label"].add(data["label"])
	print("%d perturbation edges have been added back" % n)
	return sccg

def reach(g, initial_nodes, final_nodes, inevitable):
	#To make things shorter (and easier to read if converted to image), instead of a custom 'valid' attribute, I just say that valid nodes are purple

	"""
	1. Mark initial nodes as valid
	2. Get valid nodes from the topological order
	"""

	def mark(n):
		g.node[n]["color"] = "purple"
	def is_marked(n):
		return g.node[n].get("color") == "purple"

	for n in final_nodes:
		mark(n)

	final_valid = set(final_nodes)

	def always_to_valid(n):
		nexts = [v for (_,v,data) in g.out_edges(nbunch=n,data=True) \
				if not data.get("color")]
		if not nexts:
			return False
		else:
			return final_valid.issuperset(nexts)

	def exists_to_valid(n):
		nexts = [v for (_,v,data) in g.out_edges(nbunch=n,data=True) \
				if not data.get("color")]
		for n in nexts:
			if n in final_valid:
				return True
		return False

	test_to_valid = always_to_valid if inevitable else exists_to_valid

	for node in nx.topological_sort(g, reverse=True):
		if node not in final_nodes:
			is_valid=True					#only used for inevitability
			only_unvalid_red=True				#this bool is used to check if the outgoing edges are either [] or only red edges that go to unvalid nodes (a valid red edge to a node will still work)
			for (u,v, data) in g.out_edges(nbunch=node, data=True):		#checking if all edges lead to a valid node
				v_valid=is_marked(v)
				col=data.get("color")
				if inevitable:
					if col=="red":
						if v_valid:					#if only one red edge leads to a valid state it works as well
							mark(node)
							is_valid=False
							break
					else:
						only_unvalid_red=False
						if not v_valid:					#if one black edge doesn't lead to a valid state, then the node isn't valid (yet?)
							is_valid=False
				else:
					if v_valid:
						mark(node)
						break

			if is_valid and inevitable and not only_unvalid_red:
				mark(node)

		if test_to_valid(node):
			final_valid.add(node)

	one_path_exists=False
	for node in initial_nodes:
		if is_marked(node):
		    one_path_exists = True

	return final_valid, one_path_exists


def graph_reduction(graph, initial_nodes, final_nodes, final_valid):		#reducing the graph to only the valid nodes, and then reducing even more	WARNING: may return confusing results: paths that seem inevitable but are not.
	g = graph.copy()				#copying to have both graphs
	nodes_to_del=[]
	i_nodes= initial_nodes.copy()
	for node in g.nodes():
		if "color" not in g.node[node]:
			if node in initial_nodes:
				initial_nodes.remove(node)		
			nodes_to_del.append(node)	
	for node in nodes_to_del:
		g.remove_node(node)
		if node in final_valid:
			final_valid.remove(node)
	print("%s nodes in condensed graph without non-valid nodes" % len(g))

	return g


def keep_only_subsets(s):
	to_remove=[]
	for fset in s:
		for fset2 in s:
			if fset2 != fset and fset2.issubset(fset):
				if fset not in to_remove:
					to_remove.append(fset)
	for ss in to_remove:
		s.remove(ss)

def has_superset(e, s):
	e = set(e)
	for f in s:
		if e.issuperset(f):
			return True
	return False


def get_mutations(g, initial_nodes, final_nodes):

	def enum(g,x,past,past_m,sols,last_was_red=False,follow_black=True):
		if has_superset(past_m, sols):
			return
		if x in final_nodes:
			if not follow_black or last_was_red:
				#print("found %s" % past_m)
				sols.add(past_m)
			return
		for (u,y) in g.out_edges(nbunch=x):
			if y not in past:
				if g.edge[u][y].get("color") == "red":
					#print("edge (%s,%s) is red" % (x,y))
					for l in g.edge[x][y]['label']:
						ms= frozenset([l]+list(past_m))
						enum(g,y,past+[y],ms,sols,last_was_red=True,follow_black=follow_black)
				else:
					if follow_black:
						enum(g,y, past+[y], past_m, sols,follow_black=follow_black)

	init_s=set()
	enum(g, initial_nodes[0], [initial_nodes[0]], frozenset(),init_s, follow_black=False)
	keep_only_subsets(init_s)

	print("# Computing temporal mutations")

	for x in nx.topological_sort(g, reverse=True):
		g.node[x]["muts"]=set()
		if x not in final_nodes:
			for (u,y,data) in g.out_edges(nbunch=x,data=True):
				for e in g.node[y]["muts"]:
					if data.get("color") == "red":				#if red edge, add every label to every set of the node
						for l in g.edge[x][y]['label']:
							g.node[x]["muts"].add((l,)+e)
					elif y not in final_nodes:
						g.node[x]["muts"].add(e)
		else:
			g.node[x]["muts"].add(())

	s=g.node[initial_nodes[0]]["muts"]
	s = [e for e in s if not has_superset(set(e), init_s)]
	s.sort()
	#keep_only_subsets(s)		STILL TO IMPROVE?

	return(s, init_s)

def copy_node(g, gg, node):
	gg.add_node(node)
	n=g.node[node]
	if "members" in n:
		gg.node[node]["members"]=n["members"]
	if "color" in n:
		gg.node[node]["color"]=n["color"]

def copy_edge(g, gg, u, v):
	gg.add_edge(u,v)
	e=g.edge[u][v]
	if "label" in e:
		gg.edge[u][v]["label"]=e["label"]
	if "color" in e:
		gg.edge[u][v]["color"]=e["color"]


def find_a_mutation_path(g, i_states, final_nodes, e_path):		#we want a path from an initial node to a valid node, with branchings where the path could branch

	"""
	1. Find an initial node that contain the wanted path in its muts
	2. Make a new graph with this node
	3. From this node, follow the path:
		3.1. If mutated edge following the path, add the edge and destination node to the new graph, and make the destination node the node from which the path has to be followed
			** if branching already found, break the loop
		3.2. If path can be followed by a normal edge, store the node ; store a maximum of 2 nodes
			** break the loop if a correct mutated edge have been found
		3.3. If normal edge (not improving the path), add it, and branching is found
			** break the loop if a correct mutated edge have been found
		3.4. Once all out-going edges have been searched through, add an other node if needed
		3.5. If no correct mutated edge have been found, add a node stored in 3.2, and make it the node from which the path has to be followed
	4. Repeat 3 until the path is followed entirely
	5. Return the new graph
	"""

	path=e_path
	i=0
	while path not in g.node[i_states[0]]["muts"]:
		i+=1

	gg=nx.DiGraph()
	current_node= i_states[0]
	copy_node(g, gg, current_node)

	while path:
		buff=current_node
		one_other_node=False
		valid_nodes=[]
		progression=False
		for u,v, data in g.out_edges(nbunch=current_node, data=True):
			if data.get("label")!=None:
				if path[0] in data.get("label"):
					copy_node(g,gg, v)
					copy_edge(g, gg, u, v)
					current_node=v
					progression=True
					if one_other_node or valid_nodes!=[]:
						break
			elif path in g.node[v]["muts"]:
				if len(valid_nodes)<3:
					valid_nodes.append(v)
					if progression:
						break
			elif data.get("color")!="red" and not one_other_node:
				copy_node(g,gg, v)
				copy_edge(g,gg, u, v)
				one_other_node=True
				if progression:
					break

		if progression:
			path=tuple(list(path)[1:])			#moved here to avoid errors on the last loop
			if valid_nodes!=[] and not one_other_node:
				copy_node(g, gg, valid_nodes[0])
				copy_edge(g, gg, buff, current_node)
		else:
			copy_node(g, gg, valid_nodes[0])
			copy_edge(g, gg, current_node, valid_nodes[0])
			if not one_other_node and len(valid_nodes)>1:
				copy_node(g, gg, valid_nodes[1])
				copy_edge(g, gg, current_node, valid_nodes[1])
			current_node=valid_nodes[0]

	return gg
